/*****************************************************************************\
 * AstroOps
 *
 * AstroOps  is a 'catch-all' class that performs odd handy calculations
 *
 * author: mark huss (mark@mhuss.com)
 * Based on Bill Gray's open-source code at projectpluto.com
 *
\*****************************************************************************/

#if !defined( ASTRO_OPS__H )
#define ASTRO_OPS__H

#include"MathOps.h"

// * * * * * Handy functions that didn't fit anywhere else * * * * *

class AstroOps {
public:

  enum SolExType { SPRING, SUMMER, AUTUMN, WINTER };

 /**
  * meanSolarLongitude() - calculate the mean solar latitude on the given T
  *                  (Meeus, Ch. 25)
  *
  * @param t - day to use in Julian Millenia
  *
  * @return - the longitude in radians
  */
  static double meanSolarLongitude( double t );

 /**
  * solarLongitude() - calculate the solar latitude on the given JD
  *                   (Meeus, Ch. 25)
  *
  * @param jd - day to use
  *
  * @return - the longitude in radians
  */
  static double solarLongitude( double jd );

 /**
  * meanSolarAnomaly() - calculate the mean solar anomaly on the given T
  *                     (Meeus, Ch. 25)
  *
  * @param T - day to use in Julian Centuries
  *
  * @return - the anomaly in radians
  */
  static double meanSolarAnomaly( double T ) {
    return MathOps::toRadians(357.52911 + 35999.05029*T - 0.0001537*T*T);
  }

 /**
  * earthEccentricity() - calculate the eccentricity of the Earth's orbit on
  *                       the given T (Meeus, Ch. 25)
  *
  * @param T - day to use in Julian Centuries
  *
  * @return - the eccentricity
  */
  static double earthEccentricity( double T ) {
    return 0.016708634 - 0.000042037*T - 0.0000001267*T*T;
  }

 /**
  * equationOfTime() - calculate the equation of time on the given JD
  *                   (Meeus, Ch. 28)
  *
  * @param jd - day to use
  *
  * @return - the time in minutes between apparent time and mean time
  */
  static double equationOfTime( double jd );

 /**
  * equinoxSolstice() - calculate the time of the Equinoxes and Solstices
  *
  * @param year - year to calculate quarter for
  * @param type - which solar quarter (SPRING, etc.)
  * @param local - true for local time, else UTC
  *
  * @return - the relatively exact jd time of the spec'd event
  */
  static double equinoxSolstice( int year, SolExType type, bool local = false );

 /**
  * meanObliquity()  - calculate the mean obliquity for a given jd
  *
  * @param jd - julian day (valid for years -8000 to +12000)
  *
  * @return Mean obliquity (epsilon sub 0) in radians.
  *
  * The obliquity formula (and all the magic numbers below) come from Meeus,
  * Astro Algorithms.
  *
  */
  static double meanObliquity( double jd );

 /**
  * nutation() - calculate delta phi and/or delta epsilon for the given jd
  *
  * @param - julian day
  * @param pDPhi - [out] nutation (delta phi) in arcseconds
  * @param pDEpsilon - [out] nutation (delta epsilon) in arcseconds
  *
  * The nutation formula (and all the magic numbers below) come from
  * p 132-5, Meeus,  Astro Algorithms
  *
  * Either pointer can be NULL,  in which case that value is
  * not computed.  (we added this because sometimes,  you want
  * only pDPhi or pDEpsilon;  in such cases,  computing _both_
  * is a waste of perfectly good CPU time)
  */
  static void   nutation( double jd, double* pDPhi, double* pDEpsilon );

};

#endif  /* #if !defined( ASTRO_OPS__H ) */


