/*****************************************************************************\
 * DateOps.cpp
 *
 * DateOps contains misc. time and date operations
 *
 * author: mark huss (mark@mhuss.com)
 * Based in part on Bill Gray's open-source code at projectpluto.com
 *
\*****************************************************************************/

#include "DateOps.h"
#include "DateCalendars.h"
#include "MathOps.h"

#include <string.h>
#include <stdio.h>
#include <time.h>

const double DateOps::DST_OFFSET = Astro::DAYS_PER_HOUR;

const char* DateOps::MONTH[] = { "", /* month 0 */
    "January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"
};

const char* DateOps::MONTH3[] = { "", /* month 0 */
    "Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"
};

/*----------------------------------------------------------------------------
 * greenwichSiderealTime
 *
 * @param jd - julian day
 *
 * @return apparent Greenwich sidereal time (in Radians) for the given jd
 *
 * See p 84,  in Meeus
 */
double DateOps::greenwichSiderealTime( double jd )
{
  jd -= Astro::J2000;
  double jdc = jd/Astro::DAYS_PER_CENTURY;      /* convert jd to julian centuries */
  double intPart = floor( jd );
  double fracPart = jd - intPart;
  double rval = 280.46061837 +
                360.98564736629 * fracPart +
                .98564736629 * intPart +
                jdc * jdc * ( 3.87933e-4 - jdc / 38710000. );

  return MathOps::toRadians( rval );
}

//----------------------------------------------------------------------------
/**
 * formatTime(): format a time into an HH:MM or HH:MM:SS string
 *
 * @param buf - where to put the string
 * @param dayFrac - a fractional day ( >= 0.0, < 1.0 )
 * @param doSecs - true to include seconds
 */
int DateOps::formatTime( char* buf, double dayFrac, bool doSecs ) {
  if (!buf)
    return 0;

  int h, m, s, rv = 0;
  dayToHMS(h, m, s, dayFrac);

  if ( doSecs )
    sprintf( buf, "%02d:%02d:%02d", h, m, s );
  else {
    rv = roundToNearestMinute( h, m, s );
    sprintf( buf, "%02d:%02d", h, m );
  }
  return rv;
}

/**
 * formatTime(): format a time into an HH:MM string, and write it to a FILE
 *
 * @param fp - where to write the string
 * @param dayFrac - a fractional day ( >= 0.0, < 1.0 )
 * @param doSecs - true to include seconds (HH:MM:SS)
 */
int DateOps::formatTime( FILE* fp, double dayFrac, bool doSecs ) {
  if (!fp)
    return 0;

  int h, m, s, rv = 0;
  dayToHMS(h, m, s, dayFrac);

  if ( doSecs )
    fprintf( fp, "%02d:%02d:%02d", h, m, s );
  else {
    rv = roundToNearestMinute( h, m, s );
    fprintf( fp, "%02d:%02d", h, m );
  }
  return rv;
}

//----------------------------------------------------------------------------
/**
 * formatDateTime(): format a JD into a text string
 *
 * @param buf - where to put the fomatted string, must be > 12 chars long
 * @param jd
 * @param format type (DateOps::DATE_FMT)
 */
void DateOps::formatDateTime( char* clientBuf, double jd, DATE_FMT fmt ) {
  if ( !clientBuf ) return;

  int d, m, y;
  char tbuf[16] = { 0 };

  if ( fmt >= Y_MON_D_HM )
    jd += formatTime(tbuf, jd);

  DateOps::dayToDmy( jd, d, m, y );

  switch (fmt) {
  // date only
  case Y_MON_D:
    sprintf ( clientBuf, "%04d %s %02d", y, MONTH3[m], d );
    break;
  case MON_D_Y:
    sprintf ( clientBuf, "%s %02d %04d", MONTH3[m], d, y );
    break;
  case MON_D:
    sprintf ( clientBuf, "%s %02d", MONTH3[m], d );
    break;
  case Y_M_D:
    sprintf ( clientBuf, "%04d-%%02d-%02d", y, m, d );
    break;
  case M_D_Y:
    sprintf ( clientBuf, "%02d/%02d/%04d", m, d, y );
    break;
  case M_D:
    sprintf ( clientBuf, "%%02d/%02d", m, d );
    break;
  // date + time
  case Y_MON_D_HM:
    sprintf ( clientBuf, "%04d %s %02d %s", y, MONTH3[m], d, tbuf );
    break;
  case MON_D_Y_HM:
    sprintf ( clientBuf, "%s %02d %04d %s", MONTH3[m], d, y, tbuf );
    break;
  case MON_D_HM:
    sprintf ( clientBuf, "%s %02d %s", MONTH3[m], d, tbuf );
    break;
  case Y_M_D_HM:
    sprintf ( clientBuf, "%04d-%02d-%02d %s", y, m, d, tbuf );
    break;
  case M_D_Y_HM:
    sprintf ( clientBuf, "%02d/%02d/%04d %s", m, d, y, tbuf );
    break;
  case M_D_HM:
    sprintf ( clientBuf, "%02d/%02d %s", m, d, tbuf );
    break;

  default:
    *clientBuf = 0;
  };

}

/**
 * formatDateTime(): format a JD into a text string and write it to a FILE
 *
 * @param fp - where to write the formatted string
 * @param jd  - the day to format
 * @param fmt - format type (see DateOps::DATE_FMT)
 */
void DateOps::formatDateTime( FILE* fp, double jd, DATE_FMT fmt ) {
  char buf[32];
  formatDateTime( buf, jd, fmt );
  fputs( buf, fp );
}

//----------------------------------------------------------------------------
/*
 * dmyToDay()
 *
 * Get calendar data for the current year,  including the JD of New Years
 * Day for that year.  After that, add up the days in intervening months,
 * plus the day of the month:
 */
long DateOps::dmyToDay( int day, int month, int year, CalendarType calendar )
{
  long jd = 0;
  MonthDays md;
  YearEndDays yed;

  if( 0 == DateCalendars::getCalendarData( year, yed, md, calendar ) ) {
    jd = yed[0];
    for( int i=0; i<(month-1); i++ ) {
      jd += md[i];
    }
    jd += long(day - 1);
  }
  return jd;
}


//----------------------------------------------------------------------------
/**
 * timeToDay() - convert a struct tm to Julian Day
 */
double DateOps::timeToDay( struct tm* pt )
{
  int secs = pt->tm_hour * Astro::ISECONDS_PER_HOUR +
             pt->tm_min * Astro::ISECONDS_PER_MINUTE +
             pt->tm_sec;
  return double(secs)/Astro::SECONDS_PER_DAY +
         dmyToDay( pt->tm_mday,
                   pt->tm_mon + 1,        // months are zero-based
                   pt->tm_year + 1900 );  // years from 1900
}


//----------------------------------------------------------------------------
/**
 * dayToTime(): convert a JD to "local" time_t
 *
  * @param jd - jd to convert
  *
  * @return time_t
 */
time_t DateOps::dayToTime( double jd ) {
  struct tm t;
  t.tm_isdst=-1;

  DateOps::dayToDmy( long(floor(jd)), t.tm_mday, t.tm_mon, t.tm_year );
  t.tm_mon--;
  t.tm_year -= 1900;

  dayToHMS( t.tm_hour, t.tm_min, t.tm_sec, jd);
  return mktime( &t );
}

//----------------------------------------------------------------------------
/*
 * dayToDmy()
 *
 * Estimates the year corresponding to an input JD and calls get_calendar_data();
 * for that year.  Occasionally,  it will find that the guesstimate was off;
 * in such cases,  it moves ahead or back a year and tries again.  Once it's
 * done,  jd - year_ends[0] gives the number of days since New Years Day;
 * by subtracting month_data[] values,  we quickly determine which month and
 * day of month we're in.
 */

// calendar: 0 = gregorian, 1 = julian

void DateOps::dayToDmy( long jd, int& day, int& month, int& year, CalendarType calendar )
{
  day = -1;           /* to signal an error */

  YearEndDays yed;
  MonthDays md;
  year = DateCalendars::getCalendarYear(jd, calendar);
  do {
    if( 0 != DateCalendars::getCalendarData( year, yed, md, calendar ) )
      return;

    if( yed[0] > jd)
      year--;

    if( yed[1] <= jd)
      year++;

  } while( yed[0] > jd || yed[1] <= jd );

  long currJd = yed[0];
  month = -1;
  for( int i = 0; i < 13; i++) {
    day = int( jd - currJd );
    if( day < md[i] ) {
      month = i + 1;
      day++;
      return;
    }
    currJd += long( md[i] );
  }
  return;
}

//----------------------------------------------------------------------------
// Determine U.S. DST start JD (second Sunday in March as of 2007)
//
long DateOps::dstStart(int year)
{
  long jdStart = dmyToDay( 7+1, 3, year, T_GREGORIAN );
  while ( 6 != (jdStart % 7 ) ) // Sunday
    jdStart++;

  return jdStart;
}

//----------------------------------------------------------------------------
// Determine U.S. DST end JD (first Sunday in November as of 2007)
//
long DateOps::dstEnd(int year)
{
  long jdEnd = dmyToDay( 1, 11, year, T_GREGORIAN );
  while ( 6 != (jdEnd % 7 ) ) // Sunday
    jdEnd++;

  return jdEnd;
}

//----------------------------------------------------------------------------
/**
 * tzOffsetInDays(): calculate time zone offset from Universal Time in days
 *                   using the given time_t (includes DST)
 *
 * @return Offset (-0.5 ... +0.5 )
 */
double DateOps::tzOffsetInDays(time_t tt) {
  struct tm* pt = localtime(&tt);

  int localH = pt->tm_hour;
  pt = gmtime(&tt);

  int gmtH = pt->tm_hour;
  if ( gmtH < localH )
    gmtH += Astro::IHOURS_PER_DAY;

  return hourToDay( localH - gmtH );  // local - GMT
}

//----------------------------------------------------------------------------
/**
 * dstOffsetInDays(): calculate daylight savings time offset in days at
 *                    the spec'd time
 *
 * @return Offset ( 0 or 1/24 )
 */
double DateOps::dstOffsetInDays(time_t tt) {

  struct tm* pt = localtime(&tt);
  return (pt->tm_isdst > 0) ? DST_OFFSET : 0.;
}

//----------------------------------------------------------------------------
/**
 * dayToHMS(): break the fractional part of a Julian day into hours, minutes,
 * and seconds
 *
 * @param h - where to put the hour
 * @param m - where to put the minute
 * @param s - where to put the second
 * @param dayFrac - a fractional day ( >= 0.0, < 1.0 )
 */
void DateOps::dayToHMS(int& h, int& m, int& s, double jd) {

  if (jd < 0.)
    h = m = s = 0;
  else {
    double fd = jd - floor(jd);  // get rid of whole days, if any

    fd *= Astro::HOURS_PER_DAY;
    h  = int(fd);
    fd -= h;

    fd *= Astro::MINUTES_PER_HOUR;
    m  = int(fd);
    fd -= m;

    s = int( (fd * Astro::SECONDS_PER_MINUTE) + Astro::ROUND );
  }
}

//----------------------------------------------------------------------------
/**
 * dayToHM(): break the fractional part of a Julian day into hours and minutes
 *
 * @param h - where to put the hour
 * @param m - where to put the minute
 * @param dayFrac - a Julian Day
 *
 * @return 1 if rounding wrapped to 00:00[:00], else 0
 *
 */
int DateOps::dayToHM(int& h, int& m, double jd) {
  int s;
  dayToHMS( h, m, s, jd );
  return roundToNearestMinute( h, m, s );
}

//----------------------------------------------------------------------------
/**
 * private
 * roundToNearestMinute(): round a time to the nearest minute
 *
 * if h:m wraps to zero, we roll back to 23:59 to preserve "today"
 *
 * @param h - hour
 * @param m - minute
 * @param s - second
 *
 * @return 1 if rounding wrapped to 00:00, else 0.
 */
int DateOps::roundToNearestMinute(int &h, int& m, int& s) {

  int rv = 0;
  if ( s >= Astro::ISECONDS_PER_MINUTE/2 ) { // round to nearest minute
    m++;
    if ( m >= Astro::IMINUTES_PER_HOUR ) {
      m = 0;
      h++;
      if ( h >= Astro::IHOURS_PER_DAY ) {
        h = 0;
        rv = 1;
      }
    }
  }
  return rv;
}


