/*****************************************************************************\
 * DateOps.h
 *
 * DateOps contains misc. time and date operations
 *
 * author: mark huss (mark@mhuss.com)
 * Based on Bill Gray's open-source code at projectpluto.com
 *
\*****************************************************************************/

// uncomment the following line to include an assortment of non-western
// calendars:
/* #define CALENDARS_OF_THE_WORLD */


#if !defined( DATE_OPS__H )
#define DATE_OPS__H

#include "Astro.h"
#include <time.h>

typedef int MonthDays[13];

typedef long YearEndDays[2];

class DateOps {
public:
  enum CalendarType {
    T_GREGORIAN = 0,
    T_JULIAN = 1,
    // these are only used if CALENDARS_OF_THE_WORLD is defined:
    T_HEBREW = 2,
    T_ISLAMIC = 3,
    T_REVOLUTIONARY = 4,
    T_PERSIAN = 5
  };


 /**
  * timeToDay(): convert a "local" time_t to Julian Day
  *
  * @param time_t to convert
  *
  * @return Julian Day
  */
  static double timeToLDay( time_t time ) {
    return timeToDay( localtime(&time) );
  }

 /**
  * timeToUDay(): convert UTC time_t to Julian Day
  *
  * @param time_t to convert
  *
  * @return Julian Day
  */
  static double timeToUDay( time_t time ) {
    return timeToDay( gmtime(&time) );
  }

 /**
  * now(): Determine the Julian Day value at the present moment
  *
  * @return The Julian Day value
  */
  static double now() { return timeToUDay( time(0) ); }

 /**
  * dmyToDay(): convert a day/month/year to a long Julian Day
  *
  * @param day - day of the month
  * @param month - month of the year
  * @param year - year
  * @param calendar - (optional) T_GREGORIAN or T_JULIAN, former is the default
  *
  * @return equivalent Julian Day rounded to a long
  */
  static long dmyToDay(
      int day, int month, int year, CalendarType calendar = T_GREGORIAN );

 /**
  * dmyToDoubleDay(): convert a day/month/year to a double Julian Day
  *
  * @param day - day of the month
  * @param month - month of the year
  * @param year - year
  * @param calendar - (optional) T_GREGORIAN or T_JULIAN, former is the default
  *
  * @return equivalent Julian Day as a double
  */
  static double dmyToDoubleDay(
      int day, int month, int year, CalendarType calendar = T_GREGORIAN )
  {
    return Astro::JD_GREG_DIFF + dmyToDay( day, month, year, calendar );
  }

  /*
  static double dmyToDoubleDay(
      int day, int month, int year, int second, int calendar = T_GREGORIAN )
  {
    return second/Astro::SECONDS_PER_DAY + dmyToDoubleDay( day, month, year, calendar );
  }
  */
 /**
  * dmyToDoubleDay(): convert a day/month/year to a double Julian Day
  *
  * @param day - day of the month
  * @param month - month of the year
  * @param year - year
  * @param second - second in the day ( 0 .. Astro::SECONDS_PER_DAY)
  * @param calendar - (optional) T_GREGORIAN or T_JULIAN, former is the default
  *
  * @return equivalent Julian Day as a double
  */
  static double dmyToDoubleDay(
      int day, int month, int year, int second, CalendarType calendar = T_GREGORIAN )
  {
    return second/Astro::SECONDS_PER_DAY - Astro::JD_GREG_DIFF +
        dmyToDay( day, month, year, calendar );
  }

 /**
  * dayToDmy(): convert a long Julian Day to day/month/year
  *
  * @param jd - Julian Day to convert
  * @param day& - where to put the day of the month
  * @param month& - where to put the month of the year
  * @param year& - where to put the year
  * @param calendar - (optional) T_GREGORIAN or T_JULIAN, former is the default
  */
  static void dayToDmy(
      long jd, int& day, int& month, int& year, CalendarType calendar = T_GREGORIAN );

 /**
  * dayToDmy(): convert a long Julian Day to day/month/year
  *
  * @param jd - Julian Day to convert
  * @param day* - where to put the day of the month
  * @param month* - where to put the month of the year
  * @param year* - where to put the year
  * @param calendar - (optional) T_GREGORIAN or T_JULIAN, former is the default
  */
  static void dayToDmy(
      long jd, int* day, int* month, int* year, CalendarType calendar = T_GREGORIAN )
  {
    dayToDmy( jd, *day, *month, *year, calendar );
  }

 //--------------------------------------------------------------------------
 /**
  * dstStart(): Determine the Julian Day in the specified year where Daylight
  *             Savings Time starts
  *
  * @param The year
  *
  * @return The Julian Day value
  */
  static long dstStart(int year);

 /**
  * dstEnd(): Determine the Julian Day in the specified year where Daylight
  *           Savings Time ends
  *
  * @param The year
  *
  * @return The Julian Day value
  */
  static long dstEnd(int year);

private:
  static double timeToDay( struct tm* pt );

  enum CalendarEpoch {
      E_JULIAN_GREGORIAN = 1721060L,
      // these are only used if CALENDARS_OF_THE_WORLD is defined:
      E_ISLAMIC = 1948086L,
      E_HEBREW = 347996L,
      E_REVOLUTIONARY = 2375475L
  };

  static void getJulGregYearData(
      int year, long& days, MonthDays& md, bool julian );

  static int getCalendarData(
      int year, YearEndDays& days, MonthDays& md, CalendarType calendar );

// ---------------------------------------------------------------------------
#if defined( CALENDARS_OF_THE_WORLD )

  enum {
      HALAKIM_IN_DAY = (24L * 1080L),
      JALALI_ZERO = 1947954L,
      LOWER_PERSIAN_YEAR = -1096,
      UPPER_PERSIAN_YEAR = 2327
  };

  static int mod( int x, int y );

  // Islamic calendar
  static void getIslamicYearData( int year, int& days, MonthDays& md );

  // Hebrew calendar
  static int lunationsToTishri1( int year );
  static void lunationsToDaysAndHalakim(
      int lunations, int& days, int& halakim);
  static void findTishri1( int year, int& days, int& halakim );
  static void getHebrewYearData( int year, YearEndDays& days, MonthDays& md );

  // French Revolutionary calendar
  static int jdOfFrenchRevYear( int year );
  static void getRevolutionaryYearData(
      int year, YearEndDays& days, MonthDays& md );

  // Persian (Jalaali) calendar
  static int jalaliJd0( int jalaliYear );
  static void getJalaliYearData(
      const int year, YearEndDays& days, MonthDays& md );

#endif  /* #if defined( CALENDARS_OF_THE_WORLD ) */
// ---------------------------------------------------------------------------
};

#endif  /* #if !defined( DATE_OPS__H ) */
