/*****************************************************************************\
 * Lunar.cpp
 *
 * Lunar is a class that can calculate lunar fundmentals for any reasonable
 *   time.
 *
 * author: mark huss (mark@mhuss.com)
 * Based on Bill Gray's open-source code at projectpluto.com
 *
\*****************************************************************************/

#include <stdlib.h>

#include "Lunar.h"
#include "LunarTerms.h"  // data extracted from vsop.bin file

#include "MathOps.h"
#include "DateOps.h"
#include "AstroOps.h"
#include "ObsInfo.h"
#include "Vsop.h"
#include "PlanetData.h"
#include "RiseSet.h"

//-------------------------------------------------------------------------
/**
  * calculate current phase angle in radians (Meeus' easy lower precision method)
  */
double Lunar::phaseAngle() {
  if ( !m_initialized )
    return Astro::INVALID;

  return MathOps::normalizeDeg2Rad(
      180 - MathOps::toDegrees(m_f.D)
          - 6.289 * sin( m_f.Mp )
          + 2.110 * sin( m_f.M )
          - 1.274 * sin( (2 * m_f.D) - m_f.Mp )
          - 0.658 * sin( 2 * m_f.D )
          - 0.214 * sin( 2 * m_f.Mp )
          - 0.110 * sin( m_f.D )
          );
}

//-------------------------------------------------------------------------
/**
  * Calculate age of the moon in days (0.0 to 29.53...)
  * @param jd - Julian day for which lunar age is required
  */
double Lunar::ageOfMoonInDays( Lunar& luna ) {

  // first calculate solar ecliptic longitude (in RAD)
  //
  double earthLon =
      Vsop::calcLoc( DateOps::toJulianCenturies( luna.jd() ),
                     EARTH, Vsop::ECLIPTIC_LON );
  /*
  * What we _really_ want is the location of the sun as seen from
  * the earth (geocentric view).  VSOP gives us the opposite
  * (heliocentric) view, i.e., the earth as seen from the sun.
  * To work around this, we add PI to the longitude (rotate 180 degrees)
  */
  double sunLon = earthLon + Astro::PI;

  // next calculate lunar ecliptic longitude (in RAD)
  //
  double moonLon = luna.longitudeRadians();

  // age of moon in radians = difference
  double moonAge = MathOps::normalizeRadians( Astro::RAD_PER_CIRCLE - (sunLon - moonLon) );

  // convert radians to Synodic day
  double sday = SYNODIC_MONTH * (moonAge / Astro::RAD_PER_CIRCLE);
  return sday;
}

//-------------------------------------------------------------------------
/**
  * Calculate age of the moon in days (0.0 to 29.53...)
  * @param jd - Julian day for which lunar age is required
  */
double Lunar::ageOfMoonInDays( double jd ) {

  Lunar luna( jd );
  return ageOfMoonInDays( luna );
}

//----------------------------------------------------------------------------
/**
 * Calculate the time of the quarter nearest the given Julian day, using Meeus'
 * algorithm. Input JD must be within SYNODIC_MONTH/2 days of actual event, else
 * this will return the corresponding event from another lunation.
 *
 * @param jd A Julian day close to the expected date
 * @param quarter (<TT>Lunar.NEW</TT>, <TT>Q1</TT>, <TT>FULL</TT>,
 *     or <TT>Q3</TT>)
 * @param localTZ - true to convert return value to local TZ, else UTC
 *
 * @return The relatively exact Julian day with decimal
 */
double Lunar::getPhase( long jd, Quarter quarter, bool localTZ )
{

  if (quarter < NEW)
    quarter = NEW;
  else if (quarter > Q3)
    quarter = Q3;

  bool newOrFull = ( NEW == quarter || FULL == quarter );

  double T = DateOps::toJulianCenturies( jd );
  double k = floor( T * 1236.85 ) + (double)quarter/4;

  // eccentricity of Earth's orbit around the sun
  double E = 1 - (0.002516 * T) - (0.0000074 * T * T);

  // time of *mean* phase
  double JDE = 2451550.09766
             + Lunar::SYNODIC_MONTH * k
             + 0.00015437 * T * T
             - 0.000000150 * T * T * T
             + 0.00000000073 * T * T * T * T
             + .5;

  if (localTZ)
    JDE += DateOps::tzOffsetInDays(JDE);

  // Sun's mean anomaly
  double M = getFund( PhaseFundimentals_M, k, T );
  // Moon's mean anomaly
  double Mp = getFund( PhaseFundimentals_Mp, k, T );
  // Moon's argument of latitude
  double F = getFund( PhaseFundimentals_F, k, T );
  // Longitude of the ascending node of the lunar orbit
  double Om = getFund( PhaseFundimentals_Om, k, T );

  // calculate correction to *true* phase
  const LunarTermsPh* terms = newOrFull ? LunarPhaseNF : LunarPhaseQ;

  // first group of periodic terms
  double correction = 0.;
  double sumOfPeriodicArguments;
  for( int i=0; i<N_LUNAR_TERMS_PH; i++ ) {

    sumOfPeriodicArguments = M * terms[i].m;
    sumOfPeriodicArguments += Mp * terms[i].mp;
    sumOfPeriodicArguments += F * terms[i].f;
    sumOfPeriodicArguments += Om * terms[i].om;

    double phasePeriodicTerm = sin( sumOfPeriodicArguments ) * terms[i].nm;
    for ( int e = terms[i].e; e > 0; e-- )
      phasePeriodicTerm *= E;

    correction += phasePeriodicTerm;
  }
  JDE += correction;

  if ( !newOrFull ) {
    // corrections for 1Q & 3Q only
    double W = 0.00306
             - 0.00038 * E * cos(M)
             + 0.00026 * cos(Mp)
             - 0.00002 * cos(Mp-M)
             + 0.00002 * cos(Mp+M)
             + 0.00002 * cos(F+F);

    JDE += ( Q1 == quarter) ? W : -W;
  }

  // additional corrections for all phases
  JDE += 0.000325 *
      sin( MathOps::toRadians( 299.77 +  0.107408 * k - 0.009173 * T * T ) ); // A1
  JDE += 0.000165 *
      sin( MathOps::toRadians( 251.88 +  0.016321 * k ));   // A2
  JDE += 0.000164 *
      sin( MathOps::toRadians( 251.83 + 26.651886 * k ));   // A3
  JDE += 0.000126 *
      sin( MathOps::toRadians( 349.42 + 36.412478 * k ));   // A4
  JDE += 0.000110 *
      sin( MathOps::toRadians(  84.66 + 18.206239 * k ));   // A5
  JDE += 0.000062 *
      sin( MathOps::toRadians( 141.74 + 53.303772 * k ));   // A6
  JDE += 0.000060 *
      sin( MathOps::toRadians( 207.14 +  2.453732 * k ));   // A7
  JDE += 0.000056 *
      sin( MathOps::toRadians( 154.84 +  7.306860 * k ));   // A8
  JDE += 0.000047 *
      sin( MathOps::toRadians(  34.52 + 27.261239 * k ));   // A9
  JDE += 0.000042 *
      sin( MathOps::toRadians( 207.19 +  0.121824 * k ));   // A10
  JDE += 0.000040 *
      sin( MathOps::toRadians( 291.34 +  1.844379 * k ));   // A11
  JDE += 0.000037 *
      sin( MathOps::toRadians( 161.72 + 24.198154 * k ));   // A12
  JDE += 0.000035 *
      sin( MathOps::toRadians( 239.56 + 25.513099 * k ));   // A13
  JDE += 0.000023 *
      sin( MathOps::toRadians( 331.55 +  3.592518 * k ));   // A14

  return JDE;
}

//----------------------------------------------------------------------------
 /**
  * Calculate the lunation for specified Julian day.
  * <P>
  * A "lunation" is E. W. Brown's numbered series of lunar cycles.
  * Lunation 1 was on January 16, 1923.
  *
  * @param jd Julian day
  *
  * @return Lunation number
  */
  int Lunar::lunation( double jd ) {
    int lun = (int)( ( jd - LUNATION_BASE ) / SYNODIC_MONTH );
    // "++" to make 1-based (vs. 0-based)
    return ++lun;
  }

//----------------------------------------------------------------------------
 /**
  * Return approximate JD of next new moon
  */
  double Lunar::nextMoon(double jd) {
    if (jd < 0.)
      jd = DateOps::now(true);

    int lun = lunation(jd);
    do {
      jd += 1.0;
    } while ( lunation(jd) == lun );
    return jd;
  }

//----------------------------------------------------------------------------
 /**
  * Build a current lunar information summary String.
  * <P>
  * The summary contains the lunation, age in days, rise, and set
  * times.
  *
  * @param buf - where to create the summary string
  * @param oi - The observer's location
  * @param html - true to format summary string with HTML breaks
  *
  * @return age in days
  */
  void Lunar::summary(int& days, int& lun, char* buf, ObsInfo& oi, bool doHtml) {
    if (!buf)
      return;

    const char* br = doHtml ? "<BR>" : "\n";
    const char* sp = doHtml ? "&nbsp;" : " ";

    double now = DateOps::now(true);
    lun = lunation(now);

    double age = ageOfMoonInDays(now);
    days = int( age + Astro::ROUND );
    if (days < 1 || days > 29)
      days = 29;

    int d, m, y;
    DateOps::dayToDmy( now, d, m, y );

    int d1, d2, d3, m1, m2, m3, y1, y2, y3;
    double jdNext = nextMoon(now - SYNODIC_MONTH);

    double jd1 = getPhase( jdNext, Lunar::NEW, true );
    DateOps::dayToDmy( jd1, d1, m1, y1 );

    jdNext += SYNODIC_MONTH/2;
    double jd2 = getPhase( jdNext, Lunar::FULL, true );
    DateOps::dayToDmy( jd2, d2, m2, y2 );

    jdNext += SYNODIC_MONTH/2;
    double jd3 = getPhase( jdNext, Lunar::NEW, true );
    DateOps::dayToDmy( jd3, d3, m3, y3 );

    sprintf( buf, "%sAge %2.2f days%s", sp, age, br );

    char* pBuf = buf + strlen(buf);

    double jdRS = floor(now) - Astro::JD_DIFF - DateOps::tzOffsetInDays(now);

    TimePairRS rsYesterday, rsToday, rsTomorrow;
    RiseSet::getTimes( rsToday, RiseSet::MOON, jdRS, oi );
    double rise = rsToday.rise, set = rsToday.set;

    char riseMsg[32] = { 0 };
    char riseTime[8] = { 0 };
    char setMsg[32] = { 0 };
    char setTime[8] = { 0 };
    if ( rise >= 0. && set >= 0. ) { // normal condition
      DateOps::formatTime(riseTime, rise);
      DateOps::formatTime(setTime, set);
      if ( rise < set )
        sprintf( pBuf, "%sRises %s%s"
                       "%sSets%s%s%s%s",
                 sp, riseTime, br,
                 sp, sp, sp, setTime, br);
      else
        sprintf( pBuf, "%sSets%s%s%s%s"
                       "%sRises %s%s",
                 sp, sp, sp, setTime, br,
                 sp, riseTime, br );
    }
    else {
      if ( rise < 0. ) {
        RiseSet::getTimes( rsYesterday, RiseSet::MOON, jdRS-1, oi );
        rise = rsYesterday.rise;
        riseMsg[0] = '(';
        DateOps::formatDateTime( riseMsg+1, now-1, DateOps::MON_D );
        strcat( riseMsg, ")");
      }
      if ( set < 0. ) { //|| (set < rise && 0 == *riseMsg) ) {
        RiseSet::getTimes( rsTomorrow, RiseSet::MOON, jdRS+1, oi );
        set = rsTomorrow.set;
        setMsg[0] = '(';
        DateOps::formatDateTime( setMsg+1, now+1, DateOps::MON_D );
        strcat( setMsg, ")");
      }
      DateOps::formatTime(riseTime, rise);
      DateOps::formatTime(setTime, set);

      if ( *riseTime ) {
        if ( *setTime ) {  // both
          sprintf( pBuf, "%sRises %s %s%s"
                         "%sSets%s%s%s %s%s",
                   sp, riseTime, riseMsg, br,
                   sp, sp, sp, setTime, setMsg, br);
        }
        else  // riseTime only
          sprintf( pBuf, "%sRises %s %s%s", sp, riseTime, riseMsg, br);
      }
      else if ( *setTime ) // setTime only
        sprintf( pBuf, "%sSets %s %s%s", sp, setTime, setMsg, br);
    }


    pBuf = pBuf + strlen(pBuf);

    char phase[32] = { 0 };
    if (d == d1 && m == m1) {

      DateOps::formatTime( phase, jd1 );
      sprintf( pBuf, "%sNew Moon: %s%s", sp, phase, br );
      days = 1;
    }
    else if ( d == d2 && m == m2 ) {
      DateOps::formatTime( phase, jd2 );
      sprintf( pBuf, "%sFull Moon: %s%s", sp, phase, br );
    }
    else if (d == d3 && m == m3) {
      DateOps::formatTime( phase, jd3 );
      sprintf( pBuf, "%sNew Moon: %s%s", sp, phase, br );
      days = 1;
    }

  }

//----------------------------------------------------------------------------
/**
 * getFund(): calculate an individual fundimental
 *
 * @param pTerms - points to array of doubles
 * @param t - time in decimal Julian centuries
 *
 * @return - corresponding angle in radians
 */
double Lunar::getFund( const double* pTerms, double t )
{
  double d = *pTerms++;
  double tpower = t;
  for( int i=N_LFTERM-1; i!=0; i-- ) {
    d += tpower * (*pTerms++);
    tpower *= t;
  }
  return MathOps::normalizeDeg2Rad( d );
}

//----------------------------------------------------------------------------
/**
 * getFund(): calculate an individual fundimental
 *
 * @param fundArray - points to array of doubles
 * @param k - phase constant, 0 = new moon Jan 6, 2000
 * @param t - time in decimal Julian centuries
 *
 * @return - corresponding angle in radians
 */
double Lunar::getFund( const double* fundArray, double k, double t ) {
  double d = fundArray[0] + k * fundArray[1];
  double tpow = t * t;    // tpow = T^2, T^3, ...
  for( int i=2; i<N_LUNAR_FUND_PH; i++ ) {
    d += tpow * fundArray[i];
    tpow *= t;
  }
  return MathOps::normalizeDeg2Rad( d );
}

//----------------------------------------------------------------------------
/**
 * calcFundamentals() - calculate the fundamanentals given the time
 *
 * @param t = time in decimal julian centuries
 *
 * @affects - class LunarFundimentals values, class lat/lon/rad, init flag
 */
void Lunar::calcFundamentals( double jd )
{
  m_jd = jd;
  double t = DateOps::toJulianCenturies( jd );

  m_f.Lp = getFund( LunarFundimentals_Lp, t );
  m_f.D =  getFund( LunarFundimentals_D, t );
  m_f.M =  getFund( LunarFundimentals_M, t );
  m_f.Mp = getFund( LunarFundimentals_Mp, t );
  m_f.F =  getFund( LunarFundimentals_F, t );

  m_f.A1 = MathOps::normalizeDeg2Rad( 119.75 + 131.849 * t );
  m_f.A2 = MathOps::normalizeDeg2Rad( 53.09 + 479264.290 * t );
  m_f.A3 = MathOps::normalizeDeg2Rad( 313.45 + 481266.484 * t );
  m_f.T  = t;
  m_f.T2 = t * t;

  // indicate values need to be recalculated
  m_lat = m_lon = m_r = -1.;

  // set init'd flag to true
  m_initialized = true;
}

//---------------------------------------------------------------------------
/**
 * calcAllLocs() - calculate all three location elements of the moon at the
 *                 specified time
 *
 * @param lon - where to put the longitude
 * @param lat - where to put the latitude
 * @param rad - where to put the radius
 * @param t - time to calculate location elements for
 */
void Lunar::calcAllLocs(
    double& lon,            // returned longitude
    double& lat,            // returned latitude
    double& rad,            // returned radius vector
    double jd)              // time
{
  Lunar luna(jd);
  lon = luna.longitudeRadians();
  lat = luna.latitudeRadians();
  rad = luna.radius();
}

//----------------------------------------------------------------------------
/**
 * calcLonRad(): calculate longitude and radius
 *
 * NOTE: calcFundamentals() must have been called first
 *
 */
void Lunar::calcLonRad()
{
  if ( !m_initialized ) {
    m_r = m_lon = -1.;
  }
  else {
    const LunarLonRadTerm* tptr = LunarLonRad;

    double sl = 0., sr = 0.;
    const double e = 1. - .002516 * m_f.T - .0000074 * m_f.T2;

    for( int i=N_LTERM; i!=0; i-- ) {
      if( labs( tptr->sl ) > 0 || labs( tptr->sr ) > 0 ) {
        double arg;

        switch( tptr->d )
        {
          case  1:   arg = m_f.D;          break;
          case -1:   arg =-m_f.D;          break;
          case  2:   arg = m_f.D+m_f.D;    break;
          case -2:   arg =-m_f.D-m_f.D;    break;
          case  0:   arg = 0.;             break;
          default:   arg = (double)(tptr->d) * m_f.D;  break;
        }

        switch( tptr->m )
        {
          case  1:   arg += m_f.M;         break;
          case -1:   arg -= m_f.M;         break;
          case  2:   arg += m_f.M+m_f.M;   break;
          case -2:   arg -= m_f.M+m_f.M;   break;
          case  0:           ;             break;
          default:   arg += (double)(tptr->m) * m_f.M;  break;
        }

        switch( tptr->mp )
        {
          case  1:   arg += m_f.Mp;        break;
          case -1:   arg -= m_f.Mp;        break;
          case  2:   arg += m_f.Mp+m_f.Mp; break;
          case -2:   arg -= m_f.Mp+m_f.Mp; break;
          case  0:           ;             break;
          default:   arg += (double)(tptr->mp) * m_f.Mp;  break;
        }

        switch( tptr->f )
        {
          case  1:   arg += m_f.F;         break;
          case -1:   arg -= m_f.F;         break;
          case  2:   arg += m_f.F+m_f.F;   break;
          case -2:   arg -= m_f.F+m_f.F;   break;
          case  0:           ;             break;
          default:   arg += (double)(tptr->f) * m_f.F;  break;
        }

        if( tptr->sl )
        {
          double term = (double)(tptr->sl) * sin(arg);
          for( int j=abs(tptr->m); j!=0; j-- )
            term *= e;
          sl += term;
        }

        if( tptr->sr )
        {
          double term = (double)(tptr->sr) * cos(arg);
          for( int j=abs(tptr->m); j!=0; j-- )
            term *= e;
          sr += term;
        }
      }
      tptr++;
    }

    sl += 3958. * sin( m_f.A1 ) +
          1962. * sin( m_f.Lp - m_f.F ) +
          318.  * sin( m_f.A2 );

    m_lon = (m_f.Lp * 180. / Astro::PI) + sl * 1.e-6;

    // reduce signed angle to ( 0 < m_lon < 360 )
    m_lon = MathOps::normalizeDegrees( m_lon );
    m_r = 385000.56 + sr / 1000.;
  }
}

//----------------------------------------------------------------------------
/**
 * latitude(): - calculate (or return prev. calculated) latitude
 *
 * NOTE: calcFundamentals() must have been called first
 *
 */
double Lunar::latitude()
{
  if ( !m_initialized )
    return -1.;

  if ( m_lat < 0. ) {
    const LunarLatTerm* tptr = LunarLat;
    double rval = 0.;

    const double e = 1. - .002516 * m_f.T - .0000074 * m_f.T2;

    for( int i=N_LTERM; i!=0; i-- ) {

      if( labs( tptr->sb ) > 0. ) {
        double arg;

        switch( tptr->d )
        {
          case  1:   arg = m_f.D;          break;
          case -1:   arg =-m_f.D;          break;
          case  2:   arg = m_f.D+m_f.D;    break;
          case -2:   arg =-m_f.D-m_f.D;    break;
          case  0:   arg = 0.;             break;
          default:   arg = (double)(tptr->d) * m_f.D;  break;
        }

        switch( tptr->m )
        {
          case  1:   arg += m_f.M;         break;
          case -1:   arg -= m_f.M;         break;
          case  2:   arg += m_f.M+m_f.M;   break;
          case -2:   arg -= m_f.M+m_f.M;   break;
          case  0:           ;             break;
          default:   arg += (double)(tptr->m) * m_f.M;  break;
        }

        switch( tptr->mp )
        {
          case  1:   arg += m_f.Mp;        break;
          case -1:   arg -= m_f.Mp;        break;
          case  2:   arg += m_f.Mp+m_f.Mp; break;
          case -2:   arg -= m_f.Mp+m_f.Mp; break;
          case  0:           ;             break;
          default:   arg += (double)(tptr->mp) * m_f.Mp;  break;
        }

        switch( tptr->f )
        {
          case  1:   arg += m_f.F;         break;
          case -1:   arg -= m_f.F;         break;
          case  2:   arg += m_f.F+m_f.F;   break;
          case -2:   arg -= m_f.F+m_f.F;   break;
          case  0:           ;             break;
          default:   arg += (double)(tptr->f) * m_f.F;  break;
        }

        double term = (double)(tptr->sb) * sin( arg );
        for( int j = abs(tptr->m); j!=0; j-- )
          term *= e;

        rval += term;
      }
      rval += -2235. * sin( m_f.Lp ) +
               382.  * sin( m_f.A3 ) +
               175.  * sin( m_f.A1 - m_f.F ) +
               175.  * sin( m_f.A1 + m_f.F ) +
               127.  * sin( m_f.Lp - m_f.Mp ) -
               115.  * sin( m_f.Lp + m_f.Mp );

      tptr++;
    }
    m_lat = rval * 1.e-6;
  }
  return m_lat;
}


