/**
 * LunarCalc is a class that does lunar calculations that do not
 * directly depend on the lunar fundamentals (although some do
 * need to call functions in Lunar).
 * <P>
 * The code in this class also uses other classes outside of Lunar
 * (and some of these classes in turn use Lunar), so these functions
 * are in a separate class, to avoid circular dependencies.
 * <P>
 * Based in part on C code by Bill Gray (www.projectpluto.com)
 */

#include "LunarCalc.h"

double LunarCalc::ageOfMoonInDays( double jd ) {

    double centuries = Astro.toMillenia( jd ); // convert jd to jm ref. J2000

    // first calculate solar ecliptic longitude (in RAD)
    //
    double earthLon = Vsop::calcLoc( centuries, EARTH, Vsop::ECLIPTIC_LAT );
   /*
    * What we _really_ want is the location of the sun as seen from
    * the earth (geocentric view).  VSOP gives us the opposite
    * (heliocentric) view, i.e., the earth as seen from the sun.
    * To work around this, we add PI to the longitude (rotate 180 degrees)
    */
    double sunLon = earthLon + Astro::PI;

    // next calculate lunar ecliptic longitude (in RAD)
    //
    Lunar luna = new Lunar(centuries);
    double moonLon = luna.getLongitudeRadians();
    }
    catch (NoInitException ni) {}

    // age of moon in radians = difference
    double moonAge = AstroOps.normalizeRadians( Astro.TWO_PI - (sunLon - moonLon) );

    // convert radians to Synodic day
    double sday = SYNODIC_MONTH * (moonAge / Astro.TWO_PI);

    return sday;
  }

 /**
  * Calculate the age of the moon in days for the given <TT>Calendar</TT>
  * instance.
  *
  * @param cal - <TT>java.util.GregorianCalendar</TT> holding date for
  * which lunar age is required
  *
  * @return The lunar age in days (0.0 to 29.5306)
  */
  public static double ageOfMoonInDays( GregorianCalendar cal ) {
    return
        ageOfMoonInDays(
            DateOps.dmyToDay(
                cal.get(Calendar.DATE),
                cal.get(Calendar.MONTH)+1,  // Calendar has 0-based months (!)
                cal.get(Calendar.YEAR)
            ) + AstroOps.toDays(
                  cal.get(Calendar.HOUR_OF_DAY) + TimeOps.tzOffset() )
        );
  }

 /**
  * Calculate the present age of the moon in days.
  *
  * @return The lunar age in days (0.0 to 29.5306)
  */
  public static double ageOfMoonInDays( /* now */ ) {
    return ageOfMoonInDays( new GregorianCalendar() );
  }

 /**
  * Calculate the lunation for specified Julian day.
  * <P>
  * A "lunation" is E. W. Brown's numbered series of lunar cycles.
  * Lunation 1 was on January 16, 1923.
  *
  * @param jd Julian day
  *
  * @return Lunation number
  */
  public static int lunation( long jd ) {
    int lun = (int)( ( (double)jd - LUNATION_BASE ) / SYNODIC_MONTH );
    // "++" to make 1-based (vs. 0-based)
    return ++lun;
  }

 /**
  * Calculate the lunation for specified <TT>GregorianCalendar</TT>
  * instance.
  * <P>
  * A "lunation" is E. W. Brown's numbered series of lunar cycles.
  * Lunation 1 was on January 16, 1923.
  *
  * @param cal <TT>java.util.GregorianCalendar</TT> to use
  *
  * @return Lunation number
  */
  public static int lunation( GregorianCalendar cal ) {
    return lunation(
        DateOps.dmyToDay(
            cal.get(Calendar.DATE),
            cal.get(Calendar.MONTH)+1,    // Calendar has 0-based months (!)
            cal.get(Calendar.YEAR) ) );
  }

 /**
  * Calculate the lunation at the present time.
  * <P>
  * A "lunation" is E. W. Brown's numbered series of lunar cycles.
  * Lunation 1 was on January 16, 1923.
  *
  * @return Lunation number
  */
  public static int lunation( /* now */ ) {
    return lunation( new GregorianCalendar() );
  }

 /**
  * Build a current lunar information summary String.
  * <P>
  * The summary contains the lunation, age in days, rise, and set
  * times.
  *
  * @param oi - The observer's location
  *
  * @return The summary string
  */
  public static String summary(ObsInfo oi) {
    GregorianCalendar now = new GregorianCalendar();

    double age = ageOfMoonInDays(now);
    int iDays = (int)age;
    int iHours = (int)(((age - iDays) * 24.) + Astro.ROUND_UP);

    String info =  "Lunation " + lunation(now)
        + ", age " + iDays + "d " + iHours + "h, \n";

   /*** commented out until RiseSet gets fixed ***/
    double offset = TimeOps.tzOffsetInDays() + TimeOps.dstOffsetInDays(now) + .5;
    double jd = (double)DateOps.calendarToDay(now) - offset;

    TimePair rs = RiseSet.getTimes( RiseSet.MOON, jd, oi, new PlanetData() );

    String setSpace = "";
    if ( rs.a >= 0.) {
      info += "Rises " + TimeOps.formatTime(rs.a);
      setSpace = " ";
    }

    if ( rs.b >= 0.)
      info += setSpace + "Sets " + TimeOps.formatTime(rs.b);
    /***/

    return info;
  }

 /**
  * Build a current lunar information summary String for Phila, PA,
  * USA (a shameless convenience function for dvaa.org).
  * <P>
  * The default summary contains the lunation, age in days, rise,
  * and set times for Philadelphia, PA, USA as the observer's
  * location.
  *
  * @return The summary string
  */
  public static String summaryPHL() {
    return summary( new ObsInfo( new Latitude(39.95), new Longitude(-75.16) ) )
           + " (Phila, PA)";
  }

  //-------------------------------------------------------------------------
  // calculate the difference between lunar and solar ecliptic longitudes
  // at the given time
  //
  private static double calcDiff( PlanetData pd, double jd ) {
    return pd.calcLon(Planets.LUNA, jd) - pd.calcLon(Planets.EARTH, jd);
  }
  //-------------------------------------------------------------------------
  // returns true if the two angles are NOT in the same quadrant
  //
  private static boolean quadDiff( double a1, double a2 ) {
    return AstroOps.quadrant( a1 ) != AstroOps.quadrant( a2 );
  }
  //-------------------------------------------------------------------------
 /**
  * Calculate the relatively exact time of the lunar quarter change
  * (NM, 1Q, FM, or 3Q).
  * <P>
  * Input must be less than one day before the exact change
  *
  * @param jd Approximate Julian day number of change
  *
  * @return The relatively exact time
  */
  public static double quarterChange( double jd ) {
    double longitude[] = new double[Astro.IHOURS_PER_DAY+1];    // 24 hrs + 1
    PlanetData pd = new PlanetData();

    // Compute the longitudes for each hour:
    //
    for( int i=0; i<=Astro.IHOURS_PER_DAY; i++ )
      longitude[i] = calcDiff( pd, jd + AstroOps.toDays(i) );

    double fraction = -1.;
    // Scan the hours, looking for quadrant change:
    //
    for( int i=0; i<Astro.IHOURS_PER_DAY; i++ ) {
      if ( quadDiff( longitude[i+1], longitude[i] ) ){
       // we found a quadrant change to refine

        double delta = 1.;
        int iterations = 10;
        fraction = AstroOps.toDays(i);
        double lonBase = longitude[i];
        double lonCur = lonBase;

        while( delta > .0001 && iterations > 0 ) {
          delta /= Astro.HOURS_PER_DAY;
          fraction += delta;
          System.out.println( "...frac=" + fraction + ", delta=" + delta );
          lonCur = calcDiff( pd, jd + fraction );
          if ( quadDiff( lonBase, lonCur ))
            iterations--;
        }
        break;
      }
    }
    return fraction;
  }
