/**
 * LunarCalc is a class that does lunar calculations that do not
 * directly depend on the lunar fundamentals (although some do
 * need to call functions in Lunar).
 * <P>
 * The code in this class also uses other classes outside of Lunar
 * (and some of these classes in turn use Lunar), so these functions
 * are in a separate class, to avoid circular dependencies.
 * <P>
 * Based in part on C code by Bill Gray (www.projectpluto.com)
 */
#if !defined( LUNAR_CALC__H )
#define LUNAR_CALC__H

#include <math.h>
#include "AstroOps.h"
#include "Lunar.h"

 /**
  * The base Julian day for Ernest W. Brown's numbered series of
  * lunations = 1923 Jan 17 02:41 UT.
  * <P>
  * This date has been widely quoted as "Jan 16 1923" and indeed it
  * was (in EST) at Yale University where Prof. Brown worked.
  */
static const double LUNATION_BASE = 2423436.40347;

class LunarCalc {
public:
  // All "magic numbers" are from Meeus, Astronomical Algorithms, 2ed

 /**
  * Calculate the age of the moon in days for the given Julian day.
  *
  * @param jd - Julian day for which lunar age is required
  *
  * @return The lunar age in days (0.0 to 29.5306)
  */
  public :static double ageOfMoonInDays( double jd );

 /**
  * Calculate the present age of the moon in days.
  *
  * @return The lunar age in days (0.0 to 29.5306)
  */
  //public static double ageOfMoonInDays( /* now */ ) {
  //  return ageOfMoonInDays( new GregorianCalendar() );
  //}

 /**
  * Calculate the lunation for specified Julian day.
  * <P>
  * A "lunation" is E. W. Brown's numbered series of lunar cycles.
  * Lunation 1 was on January 16, 1923.
  *
  * @param jd Julian day
  *
  * @return Lunation number
  */
  public :static int lunation( long jd ) {
    int lun = (int)( ( (double)jd - LUNATION_BASE ) / Lunar::SYNODIC_MONTH );
    // "++" to make 1-based (vs. 0-based)
    return ++lun;
  }

 /**
  * Calculate the lunation at the present time.
  * <P>
  * A "lunation" is E. W. Brown's numbered series of lunar cycles.
  * Lunation 1 was on January 16, 1923.
  *
  * @return Lunation number
  */
  //public static int lunation( /* now */ ) {
  //  return lunation( new GregorianCalendar() );
  //}

 /**
  * Build a current lunar information summary String.
  * <P>
  * The summary contains the lunation, age in days, rise, and set
  * times.
  *
  * @param oi - The observer's location
  *
  * @return The summary string
  */
  public :static String summary(ObsInfo& oi); {

 /**
  * Build a current lunar information summary String for Phila, PA,
  * USA (a shameless convenience function for dvaa.org).
  * <P>
  * The default summary contains the lunation, age in days, rise,
  * and set times for Philadelphia, PA, USA as the observer's
  * location.
  *
  * @return The summary string
  */
  public :static String summaryPHL();


 #endif  /* #if !defined( LUNAR_CALC__H ) */
