/*****************************************************************************\
 * MathOps.cpp
 *
 * MathOps contains misc. vector and matrix operations
 *
 * author: mark huss (mark@mhuss.com)
 * Based on Bill Gray's open-source code at projectpluto.com
 *
\*****************************************************************************/

#include "MathOps.h"

/**
 * rotateVector - rotate a vector along the spec'd axis
 *
 * @param v - vector to rotate
 * @param angle - angle of rotation
 * @param axis - axis to rotate along
 */
void MathOps::rotateVector( AstroVector& v, double angle, int axis )
{
  const double sinAng = sin( angle );
  const double cosAng = cos( angle );
  const int a = (axis + 1) % 3;
  const int b = (axis + 2) % 3;

  double temp = v[a] * cosAng - v[b] * sinAng;
  v[b] = v[b] * cosAng + v[a] * sinAng;
  v[a] = temp;
}

/**
 * polar3ToCartesian - convert polar coordinates to Cartesian
 *
 * @param v - vector to put Cartesian Coordinates into
 * @param lon - polar longitude
 * @param lat - polar latitude
 */
void MathOps::polar3ToCartesian( AstroVector& v, double lon, double lat )
{
  const double cosLat = cos( lat );

  v[0] = cos( lon ) * cosLat;
  v[1] = sin( lon ) * cosLat;
  v[2] = sin( lat );
}

/**
 * invertOrthonormalMatrix
 *
 * Inverting an orthonormal matrix is simple:
 * swap a's, b's and c's as shown below
 * This
 *    x a b
 *    a x c
 *    b c x
 *
 * @param m - matrix to invert
 */
void MathOps::invertOrthonormalMatrix( AstroMatrix& m )
{
   double temp;

   swapd( m[1], m[3], temp );
   swapd( m[2], m[6], temp );
   swapd( m[5], m[7], temp );
}

/**
 * preSpinMatrix - partial spin?
 *
 * @param m1 - first matrix
 * @param m2 - second matrix
 * @param angle
 *
 */
void MathOps::preSpinMatrix( AstroMatrix& m1, AstroMatrix& m2, double angle )
{
   const double sinAng = sin( angle );
   const double cosAng = cos( angle );
   double temp;

   for( int i=0; i<9; i+=3)
   {
      temp  = m1[i] * cosAng - m2[i] * sinAng;
      m2[i] = m2[i] * cosAng + m1[i] * sinAng;
      m1[i] = temp;
   }
}

/**
 * SpinMatrix
 *
 * @param m1 - first matrix
 * @param m2 - second matrix
 * @param angle
 *
 */
void MathOps::spinMatrix( AstroMatrix& m1, AstroMatrix& m2, double angle )
{
   const double sinAng = sin( angle );
   const double cosAng = cos( angle );
   double tval;

   for( int i=0; i<3; i++ )
   {
      tval  = m1[i] * cosAng - m2[i] * sinAng;
      m2[i] = m2[i] * cosAng + m1[i] * sinAng;
      m1[i] = tval;
   }
}

