/*****************************************************************************\
 * MathOps.h
 *
 * MathOps contains misc. vector and matrix operations
 *
 * author: mark huss (mark@mhuss.com)
 * Based on Bill Gray's open-source code at projectpluto.com
 *
\*****************************************************************************/

#if !defined( MATH_OPS__H )
#define MATH_OPS__H

#include <math.h>
#include "Astro.h"

typedef double AstroVector[3];
typedef double AstroMatrix[9];

class MathOps {

public:

 /**
  * toRadians(): convert degrees to radians
  *
  * @param an angle in degrees
  *
  * @return an angle in radians
  */
  static double toRadians( double deg ) { return deg * Astro::RADS_PER_DEGREE; }

 /**
  * secToRadians(): convert arcseconds to radians
  *
  * @param an angle in arcseconds
  *
  * @return an angle in radians
  */
  static double secToRadians( double sec ) {
    return (sec/Astro::SECONDS_PER_DEGREE) * Astro::RADS_PER_DEGREE;
  }

 /**
  * toDegrees(): convert radians to degrees
  *
  * @param an angle in radians
  *
  * @return an angle in degrees
  */
  static double toDegrees( double rad ) { return rad / Astro::RADS_PER_DEGREE; }

 /**
  * normalizeDegrees(): reduce an angle in degrees to the range (0 <= d < 360)
  *
  * @param degrees
  *
  * @return degrees
  */
  static double normalizeDegrees( double d )
  {
    d = fmod( d, Astro::DEG_PER_CIRCLE);
    return ( d < 0.) ? d + Astro::DEG_PER_CIRCLE : d;
  }

 /**
  * normalizeRadians(): reduce an angle in radians to the range (0 <= r < 2PI)
  *
  * @param radians
  *
  * @return radians
  */
  static double normalizeRadians( double r )
  {
    r = fmod( r, Astro::RAD_PER_CIRCLE );
    return ( r < 0. ) ? r + Astro::RAD_PER_CIRCLE : r;
  }

  // reduce (0 < d < 360) a positive angle in degreesand convert to radians
  //
 /**
  * normalizeDeg2Rad(): reduce an angle in degrees to the range (0 <= d < 360),
  *                     then convert to radians
  *
  * @param degrees
  *
  * @return radians
  */
  static double normalizeDeg2Rad( double d ) {
    return normalizeDegrees( d ) * Astro::RADS_PER_DEGREE;
  }

 /**
  *  quadrant(): returns the quadrant ( 0, 1, 2, or 3 ) of the spec'd angle
  *
  * @param angle - in radians
  *
  * @return quadrant of angle
  */

  static int quadrant( double angle ) {
    return (int)( normalizeRadians( angle ) * Astro::TWO_OVER_PI );
  }

 /**
  * acose(): "safe" acos which prevents overflow errors
  *
  * @param angle
  *
  * @return acos (0 ... PI)
  */
  static double acose( double angle ) {
    if( angle >= 1. )
      return( 0. );
    else if( angle <= -1. )
      return( Astro::PI );
    else
      return( acos( angle ) );
  }

 /**
  * asine(): "safe" asine which prevents overflow errors
  *
  * @param angle
  *
  * @return asin (PI/2 ... -PI/2)
  */
  static double asine( double angle ) {
    if( angle >= 1. )
      return( Astro::PI_OVER_TWO );
    else if( angle <= -1. )
      return( -Astro::PI_OVER_TWO );
    else
      return( asin( angle ) );
  }

 /**
  * setIdentityMatrix - preload a matrix:
  *    0 1 1
  *    1 0 1
  *    1 1 0
  *
  * @param m - matrix to initialize
  *
  */

 static void setIdentityMatrix( AstroMatrix m ) {
    for( int i=0; i<9; i++ )
      m[i] = ( (i & 3) ? 0. : 1. );
  }

 /**
  * swapd - inline swap ( a <--> b )
  *
  * tmp is only there to avoid a local var
  *
  * @param a - first double
  * @param b - second double
  * @param a - scratch double
  *
  */
  static void swapd( double& a, double& b, double tmp )
      { tmp = a;  a = b;  b = tmp; }

 /**
  * rotateVector - rotate a vector along the spec'd axis
  *
  * @param v - vector to rotate
  * @param angle - angle of rotation
  * @param axis - axis to rotate along
  */
  static void rotateVector( AstroVector& v, double angle, int axis );

 /**
  * polar3ToCartesian - convert polar coordinates to Cartesian
  *
  * @param v - vector to put Cartesian Coordinates into
  * @param lon - polar longitude
  * @param lat - polar latitude
  */
  static void polar3ToCartesian( AstroVector& v, double lon, double lat );

 /**
  * invertOrthonormalMatrix
  *
  * Inverting an orthonormal matrix is simple:
  * swap a's, b's and c's as shown below
  * This
  *    x a b
  *    a x c
  *    b c x
  *
  * @param m - matrix to invert
  */
  static void invertOrthonormalMatrix( AstroMatrix& m );

  static void preSpinMatrix( AstroMatrix& m1, AstroMatrix& m2, double angle );

  static void spinMatrix( AstroMatrix& m1, AstroMatrix& m2, double angle );
};

#endif  /* #if !defined( MATH_OPS__H ) */
